from __future__ import annotations

import logging
import os
import pathlib
import subprocess
from unittest import mock

import pytest

import zocalo.configuration

sample_configuration = """
version: 1

graylog:
  plugin: graylog
  protocol: UDP
  host: localhost
  port: 12201

constants:
  plugin: storage
  order:
    - cream
    - jam
  oxford_comma: undecided

sane-constants:
  plugin: storage
  oxford_comma: yes
  units: metric

environments:
  default: live
  live:
    plugins:
      - constants
      - sane-constants
  partial:
    plugins:
      - constants
  alias: partial
  part-2:
    - sane-constants
  empty: {}
"""


def _assert_configuration_is_empty(zc):
    assert zc.environments == frozenset({})
    assert "0 environments" in str(zc)
    assert "0 plugin configurations" in str(zc)


def test_return_empty_configuration_if_no_path_specified():
    with mock.patch.dict(os.environ, {"ZOCALO_CONFIG": ""}):
        zc = zocalo.configuration.from_file()
    _assert_configuration_is_empty(zc)


def test_loading_minimal_valid_configuration():
    zc = zocalo.configuration.from_string("version: 1")
    _assert_configuration_is_empty(zc)


def test_cannot_load_unversioned_yaml_files():
    with pytest.raises(zocalo.ConfigurationError, match="Invalid configuration"):
        zocalo.configuration.from_string("value: 1")


def test_cannot_load_unknown_configuration_file_versions():
    with pytest.raises(zocalo.ConfigurationError, match="not understand"):
        zocalo.configuration.from_string("version: 0")


def test_loading_minimal_valid_configuration_from_file(tmp_path):
    config_file = tmp_path.joinpath("config.yml")
    config_file.write_text("version: 1")
    zc = zocalo.configuration.from_file(os.fspath(config_file))
    _assert_configuration_is_empty(zc)
    zc = zocalo.configuration.from_file(config_file)
    _assert_configuration_is_empty(zc)
    with mock.patch.dict(os.environ, {"ZOCALO_CONFIG": os.fspath(config_file)}):
        zc = zocalo.configuration.from_file()
    _assert_configuration_is_empty(zc)


def test_cannot_load_missing_file(tmp_path):
    with pytest.raises(zocalo.ConfigurationError, match="not found"):
        zocalo.configuration.from_file(tmp_path / "missing.yml")


def test_cannot_load_invalid_file(tmp_path):
    config = tmp_path / "invalid.yml"
    config.write_text("x: y: z:")
    with pytest.raises(zocalo.ConfigurationError, match="invalid.yml"):
        zocalo.configuration.from_file(config)


def test_loading_sample_configuration():
    zc = zocalo.configuration.from_string(sample_configuration)

    assert zc.environments == frozenset(
        {"live", "partial", "part-2", "empty", "alias", "default"}
    )
    assert "6 environments" in str(zc)
    assert "3 plugin configurations" in str(zc)


def test_cannot_load_inconsistent_configuration():
    with pytest.raises(zocalo.ConfigurationError):
        zocalo.configuration.from_string(
            """
            version: 1
            environments:
              failure:
                - unreferenced-plugin
            """
        )


def test_detect_circular_aliasing_in_environment_configuration():
    with pytest.raises(zocalo.ConfigurationError, match="circular"):
        zocalo.configuration.from_string(
            """
            version: 1
            environments:
              broken: circular
              circular: broken
            """
        )


def test_detect_undefined_alias_target_in_environment_configuration():
    with pytest.raises(zocalo.ConfigurationError, match="undefined"):
        zocalo.configuration.from_string(
            """
            version: 1
            environments:
              broken: unresolvable-alias
            """
        )


def test_cannot_load_configuration_where_environments_specifies_plugin_as_string():
    with pytest.raises(zocalo.ConfigurationError, match="invalid-spec"):
        zocalo.configuration.from_string(
            """
            version: 1
            environments:
              invalid-spec: constants
            constants:
              plugin: storage
            """
        )


def test_cannot_activate_missing_environment():
    zc = zocalo.configuration.from_string("version: 1")
    with pytest.raises(ValueError):
        zc.activate_environment("live")
    assert zc.active_environments == ()
    assert "live" not in str(zc)


def test_activate_an_aliased_environment():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("default")
    assert zc.active_environments == ("default",)
    assert "default" in str(zc)


def test_activate_an_empty_environment():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("empty")
    assert zc.active_environments == ("empty",)
    assert "empty" in str(zc)


def test_activate_one_environment():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("live")
    assert zc.active_environments == ("live",)
    with pytest.raises(AttributeError):
        zc.active_environments = ("this-should-not-be-writeable",)
    assert "live" in str(zc)


def test_activate_two_environments():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("partial")
    zc.activate_environment("part-2")
    assert zc.active_environments == ("partial", "part-2")
    assert "partial" in str(zc)
    assert "part-2" in str(zc)


def test_activate_multiple_environments():
    zc = zocalo.configuration.from_string(sample_configuration)
    e = zc.activate(envs=["partial", "part-2"])
    assert e == ("partial", "part-2")
    assert zc.active_environments == ("partial", "part-2")
    assert "partial" in str(zc)
    assert "part-2" in str(zc)


def test_activate_additional_environments():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("default")
    e = zc.activate(envs=["partial", "part-2"])
    assert e == ("partial", "part-2")
    assert zc.active_environments == ("default", "partial", "part-2")
    assert "partial" in str(zc)
    assert "part-2" in str(zc)


def test_activate_default_environment():
    zc = zocalo.configuration.from_string(sample_configuration)
    e = zc.activate([])
    assert e == ("default",)
    assert zc.active_environments == ("default",)
    assert "default" in str(zc)


def test_activate_call_honours_default_flag():
    zc = zocalo.configuration.from_string(sample_configuration)
    e = zc.activate([], default=False)
    assert e == ()
    assert zc.active_environments == ()
    assert "default" not in str(zc)


def test_activate_call_works_without_default_environment():
    zc = zocalo.configuration.from_string("version: 1")
    e = zc.activate([])
    assert e == ()
    assert zc.active_environments == ()


@pytest.mark.parametrize("name", ("include", "environments", "version"))
def test_environment_can_not_reference_reserved_name(name):
    with pytest.raises(zocalo.ConfigurationError, match="reserved"):
        zocalo.configuration.from_string(
            f"""
            version: 1
            environments:
              test:
                - {name}
            """
        )


def test_unknown_plugin_definition_triggers_a_warning(caplog):
    unique_plugin_name = "testcase_for_an_unknown_plugin"
    with caplog.at_level(logging.WARNING):
        zc = zocalo.configuration.from_string(
            f"""
            version: 1
            undefined-plugin:
              plugin: {unique_plugin_name}
            """
        )
    assert unique_plugin_name in caplog.text
    assert not hasattr(zc, unique_plugin_name)


def test_configuration_can_specify_a_missing_resolution_file(tmp_path):
    zocalo.configuration.from_string(
        f"""
        version: 1
        unused-plugin:
          {tmp_path / 'missing_file'}
        """
    )


def test_configuration_can_specify_an_unreadable_resolution_file(tmp_path):
    forbidden_file = tmp_path / "forbidden_file"
    forbidden_file.write_text("This should not be accessible")
    zc = zocalo.configuration.from_string(
        f"""
        version: 1
        forbidden-plugin:
          {forbidden_file}
        environments:
         forbidden:
           - forbidden-plugin
        """
    )
    try:
        if os.name == "nt":
            subprocess.run(
                ("icacls", os.fspath(forbidden_file), "/deny", "Everyone:(R)"),
                check=True,
            )
        else:
            forbidden_file.chmod(0o000)
        with pytest.raises(PermissionError):
            zc.activate_environment("forbidden")
    finally:
        if os.name == "nt":
            subprocess.run(
                ("icacls", os.fspath(forbidden_file), "/remove:d", "Everyone"),
                check=True,
            )
        else:
            forbidden_file.chmod(0o664)


def test_plugins_can_be_configured_in_an_external_file(tmp_path):
    external_plugin = tmp_path / "external.yml"
    external_plugin.write_text(
        """
        plugin: storage
        value: sentinel
        """
    )
    zc = zocalo.configuration.from_string(
        f"""
        version: 1
        external:
          {tmp_path / 'external.yml'}
        environments:
          ext:
            - external
        """
    )
    assert "1 of" in str(zc)
    zc.activate_environment("ext")
    assert "1 of" not in str(zc)
    assert zc.storage["value"] == "sentinel"


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_loading_modular_configuration_from_string(tmp_path):
    secondary_file = tmp_path / "config.yml"
    secondary_file.write_text(sample_configuration)

    zc = zocalo.configuration.from_string(
        f"""
        version: 1
        include:
          - {secondary_file}
        """
    )
    assert zc.environments


@pytest.mark.xfail(raises=NotImplementedError, strict=True)
def test_loading_modular_configuration_from_file(tmp_path):
    secondary_file = tmp_path / "config.yml"
    secondary_file.write_text(sample_configuration)
    primary_file = tmp_path / "primary.yml"
    primary_file.write_text(
        """
        version: 1
        include:
          - config.yml
        """
    )

    zc = zocalo.configuration.from_file(primary_file)
    assert zc.environments


def test_cannot_load_modular_configuration_with_missing_reference(tmp_path):
    secondary_file = tmp_path / "non-existing-file.yml"

    with pytest.raises(FileNotFoundError):
        zocalo.configuration.from_string(
            f"""
            version: 1
            include:
              - {secondary_file}
            """
        )


def test_cannot_load_modular_configuration_with_broken_reference(tmp_path):
    secondary_file = tmp_path / "invalid.yml"
    secondary_file.write_text("x: y: z:")
    with pytest.raises(zocalo.ConfigurationError, match="invalid.yml"):
        zocalo.configuration.from_string(
            f"""
            version: 1
            include:
              - {secondary_file}
            """
        )


def test_resolve_external_references_into_home_directory():
    merged = zocalo.configuration._merge_configuration(
        """
        version: 1
        foo: ~/bar.yml
        """,
        None,
        pathlib.Path.cwd(),
    )
    assert merged == {
        "version": 1,
        "foo": pathlib.Path("~/bar.yml").expanduser(),
        "environments": {},
    }



from __future__ import annotations

import functools
import inspect
import itertools
import keyword
import logging
import operator
import os
import pathlib
import typing

import marshmallow as mm
import pkg_resources
import yaml

import zocalo.configuration.argparse
from zocalo import ConfigurationError

logger = logging.getLogger("zocalo.configuration")


class ConfigSchema(mm.Schema):
    version = mm.fields.Int(required=True)
    environments = mm.fields.Dict(
        keys=mm.fields.Str(),
        values=mm.fields.Dict(
            keys=mm.fields.Str(), values=mm.fields.List(mm.fields.Str())
        ),
    )
    include = mm.fields.List(mm.fields.Str())


class PluginSchema(mm.Schema):
    plugin = mm.fields.Str(
        required=True,
        error_messages={"required": "definition lacks the mandatory 'plugin' key"},
    )


_reserved_names = {"activated", "environments", "plugin_configurations"}


def _check_valid_plugin_name(name: str) -> bool:
    valid = (
        name.isidentifier()
        and not keyword.iskeyword(name)
        and name not in _reserved_names
    )
    if not valid:
        logger.warning(
            f"Zocalo configuration plugin '{name}' is not a valid plugin name"
        )
    return valid


_configuration_plugins = {
    e.name: e
    for e in pkg_resources.iter_entry_points("zocalo.configuration.plugins")
    if _check_valid_plugin_name(e.name)
}


@functools.lru_cache(maxsize=None)
def _load_plugin(name: str):
    if name not in _configuration_plugins:
        logger.warning(f"Zocalo configuration plugin '{name}' missing")
        return False
    return _configuration_plugins[name].load()


class Configuration:
    __slots__ = tuple(
        ["_" + name for name in _configuration_plugins]
        + ["_" + name for name in _reserved_names]
        + ["environment_cmd_args"]
    )

    def __init__(self, yaml_dict: dict):
        self._activated: typing.List[str] = []
        self._environments: typing.Dict[str, typing.List[str]] = yaml_dict.get(
            "environments", {}
        )
        self._plugin_configurations: typing.Dict[
            str, typing.Union[pathlib.Path, typing.Dict[str, typing.Any]]
        ] = {
            name: config
            for name, config in yaml_dict.items()
            if name not in ConfigSchema().fields
        }
        for name in _configuration_plugins:
            setattr(self, "_" + name, None)
        self.environment_cmd_args: typing.Tuple[str, ...] = ("-e", "--environment")

    @property
    def environments(self) -> typing.FrozenSet[str]:
        return frozenset(self._environments)

    @property
    def active_environments(self) -> typing.Tuple[str, ...]:
        return tuple(self._activated)

    def _resolve(self, plugin_configuration: str):
        try:
            configuration = self._plugin_configurations[
                plugin_configuration
            ].read_text()  # type: ignore
        except PermissionError as e:
            raise PermissionError(
                f"Plugin configuration {plugin_configuration} could not be resolved, "
                f"{e}"
            ) from None
        try:
            yaml_dict = yaml.safe_load(configuration)
        except yaml.MarkedYAMLError as e:
            raise ValueError(
                f"Plugin configuration {plugin_configuration} could not be resolved, "
                f"could not read {self._plugin_configurations[plugin_configuration]}: {e}"
            ) from None
        if not isinstance(yaml_dict, dict) or not yaml_dict.get("plugin"):
            raise ConfigurationError(
                f"Error reading configuration for plugin {plugin_configuration}: "
                f"Configuration file {self._plugin_configurations[plugin_configuration]} "
                "is missing a plugin specification"
            )
        self._plugin_configurations[plugin_configuration] = yaml_dict

    def activate_environment(self, name: str):
        """Load all plugins for a given environment."""
        if name not in self._environments:
            raise ValueError(f"Environment '{name}' is not defined")
        for config_name in self._environments[name]:
            logger.debug("Loading plugin configuration %s", config_name)
            if isinstance(self._plugin_configurations[config_name], pathlib.Path):
                self._resolve(config_name)
            configuration = self._plugin_configurations[config_name]
            plugin = _load_plugin(configuration["plugin"])  # type: ignore
            if plugin:
                plugin_parameters = inspect.signature(plugin.activate).parameters
                arguments = {"configuration": configuration, "config_object": self}
                if not any(
                    p.kind == inspect.Parameter.VAR_KEYWORD
                    for p in plugin_parameters.values()
                ):
                    arguments = {
                        p: arguments[p]
                        for p in set(arguments).intersection(plugin_parameters)
                    }
                return_value = plugin.activate(**arguments)
                setattr(self, "_" + configuration["plugin"], return_value)  # type: ignore
        self._activated.append(name)

    def activate(
        self,
        envs: typing.Optional[typing.Iterable[str]] = None,
        *,
        default: bool = True,
    ) -> typing.Tuple[str, ...]:
        """
        Activate a list of environments in order.

        :param envs: List of environments to activate. If no list is passed,
                     attempt to infer the environments from command line arguments.
        :param default: Attempt to activate environment named 'default' if no
                        environments are specified or can be inferred.
        :return: Tuple of environments activated by this function call.
        """
        if envs is None:
            envs = zocalo.configuration.argparse.get_specified_environments(
                arguments=self.environment_cmd_args
            )
        if default and not envs and "default" in self._environments:
            envs = ["default"]
        for environment in envs:
            self.activate_environment(environment)
        return tuple(envs)

    def add_command_line_options(self, parser):
        """function to inject command line parameters"""
        if "add_argument" in dir(parser):
            parser.add_argument(
                *self.environment_cmd_args,
                dest="environment",
                metavar="ENV",
                action="append",
                default=[],
                choices=sorted(self._environments),
                help="Enable site-specific settings. Choices are: "
                + ", ".join(sorted(self._environments)),
            )
        else:
            parser.add_option(
                *self.environment_cmd_args,
                dest="environment",
                metavar="ENV",
                action="append",
                default=[],
                type="choice",
                choices=sorted(self._environments),
                help="Enable site-specific settings. Choices are: "
                + ", ".join(sorted(set(self._environments) - {"default"})),
            )

    if typing.TYPE_CHECKING:
        # The configuration object will offer access to plugin objects as attributes.
        # Type checking tools will not be able to determine the type of attributes,
        # however we need to tell type checkers that accessing statically undefined
        # attributes is allowed. We do this by setting a return type on __getattr__,
        # but only when in a type checking run, as to not affect the runtime class
        # behaviour.
        def __getattr__(self, name: str) -> typing.Any:
            ...

    def __str__(self):
        environments = len(self._environments)
        plugin_configurations = len(self._plugin_configurations)
        unresolved = sum(
            1
            for conf in self._plugin_configurations.values()
            if isinstance(conf, pathlib.Path)
        )
        if unresolved:
            unresolved = f", {unresolved} of which are unresolved"
        else:
            unresolved = ""
        if not self._activated:
            activated = ""
        elif len(self._activated) == 1:
            activated = f", environment '{self._activated[0]}' activated"
        else:
            activated = f", environments {self._activated} activated"
        return (
            f"<ZocaloConfiguration containing {environments} environments,"
            f" {plugin_configurations} plugin configurations{unresolved}{activated}>"
        )

    __repr__ = __str__


for _plugin in _configuration_plugins:
    if hasattr(Configuration, _plugin):
        logger.warning(
            f"Zocalo configuration plugin '{_plugin}' is not a valid plugin name"
        )
    else:
        setattr(Configuration, _plugin, property(operator.attrgetter("_" + _plugin)))
del _plugin


def _read_configuration_yaml(configuration: str) -> dict:
    yaml_dict = yaml.safe_load(configuration)

    if not isinstance(yaml_dict, dict) or "version" not in yaml_dict:
        raise ConfigurationError("Invalid configuration specified")
    if yaml_dict["version"] != 1:
        raise ConfigurationError(
            f"This version of Zocalo does not understand v{yaml_dict['version']} configurations"
        )

    # Convert environment lists to dictionaries
    # Convert individual plugin configurations within environments to single element lists
    for environment in yaml_dict.setdefault("environments", {}):
        if isinstance(yaml_dict["environments"][environment], str):
            # Environment is an alias to another environment. Ensure the target exists.
            aliased_env = yaml_dict["environments"][environment]
            if aliased_env not in yaml_dict["environments"]:
                raise ConfigurationError(
                    f"Invalid YAML configuration: Environment {environment} aliases undefined environment {aliased_env}"
                )
            continue  # alias will be resolved after this loop
        elif isinstance(yaml_dict["environments"][environment], dict):
            pass
        elif isinstance(yaml_dict["environments"][environment], list):
            yaml_dict["environments"][environment] = {
                "plugins": yaml_dict["environments"][environment]
            }
        else:
            raise ConfigurationError(
                f"Invalid YAML configuration: Environment {environment} is not a list or dictionary"
            )
        for group in yaml_dict["environments"][environment]:
            if isinstance(yaml_dict["environments"][environment][group], list):
                pass
            elif isinstance(yaml_dict["environments"][environment][group], str):
                yaml_dict["environments"][environment][group] = [
                    yaml_dict["environments"][environment][group]
                ]
            else:
                raise ConfigurationError(
                    f"Invalid YAML configuration: Environment {environment} contains group {group} which is not a string or a list"
                )
    # Resolve environment aliases
    environment_aliases = {
        environment
        for environment in yaml_dict["environments"]
        if isinstance(yaml_dict["environments"][environment], str)
    }
    while environment_aliases:
        for environment in environment_aliases:
            aliased_env = yaml_dict["environments"][environment]
            if isinstance(yaml_dict["environments"][aliased_env], str):
                # This environment links to an alias. Skip for now.
                continue
            yaml_dict["environments"][environment] = yaml_dict["environments"][
                aliased_env
            ]
            environment_aliases.remove(environment)
            break
        else:
            raise ConfigurationError(
                f"Invalid YAML configuration: circular environment definitions for {environment_aliases}"
            )

    plugin_fields = {}
    for key in yaml_dict:
        if key in ConfigSchema().fields:
            continue
        plugin_fields[key] = mm.fields.Nested(
            PluginSchema, unknown=mm.EXCLUDE, required=True
        )
        if isinstance(yaml_dict[key], str):
            plugin_fields[key] = mm.fields.Str()
        elif isinstance(yaml_dict[key], dict) and isinstance(
            yaml_dict[key].get("plugin"), str
        ):
            plugin = _load_plugin(yaml_dict[key]["plugin"])
            if plugin and hasattr(plugin, "Schema"):
                plugin_fields[key] = mm.fields.Nested(
                    plugin.Schema, unknown=mm.EXCLUDE, required=True
                )

    class _ConfigSchema(ConfigSchema):
        class Meta:
            include = plugin_fields

    schema = _ConfigSchema()
    try:
        schema.load(yaml_dict, unknown=mm.RAISE)
    except mm.ValidationError as e:
        raise ConfigurationError(f"Invalid YAML configuration: {e}") from None

    return yaml_dict


@functools.lru_cache(maxsize=None)
def _merge_configuration(
    configuration: typing.Optional[str],
    file_: typing.Optional[pathlib.Path],
    context: pathlib.Path,
) -> dict:
    # Parse a passed YAML string or specified YAML file
    parsed_files = set()
    if file_:
        if not file_.is_file():
            raise ConfigurationError(f"Zocalo configuration file {file_} not found")
        configuration = file_.read_text()
        try:
            parsed = _read_configuration_yaml(configuration)
        except (ConfigurationError, yaml.MarkedYAMLError) as e:
            raise ConfigurationError(
                f"Error reading configuration file {file_}: {e}"
            ) from None
        parsed_files.add(file_)
    elif isinstance(configuration, str):
        parsed = _read_configuration_yaml(configuration)
    else:
        raise TypeError(
            "Either a configuration string or a configuration file must be specified"
        )

    # Resolve all lazy file references relative to the specified context parameter
    for plugin in parsed:
        if isinstance(parsed[plugin], str):
            parsed[plugin] = (
                context / pathlib.Path(parsed[plugin]).expanduser()
            ).resolve()

    # Recursively identify and merge external files (DFS)
    if parsed.get("include"):
        for include_file in parsed["include"]:
            try:
                file_reference = context.joinpath(include_file).resolve()
                include = _read_configuration_yaml(file_reference.read_text())
                assert include
            except (ConfigurationError, yaml.MarkedYAMLError) as e:
                raise ConfigurationError(
                    f"Error reading configuration file {file_reference}: {e}"
                ) from None
        raise NotImplementedError("Importing configurations is not yet supported")

    # Flatten the data structure for each environment to a deduplicated ordered list of plugins
    for environment in parsed["environments"]:
        # First, order groups alphabetically - except 'plugins', which always comes last
        ordered_plugins = [
            parsed["environments"][environment][group]
            for group in sorted(parsed["environments"][environment])
            if group != "plugins"
        ]
        ordered_plugins.append(parsed["environments"][environment].get("plugins", []))

        # Flatten the individual lists and discard duplicates
        parsed["environments"][environment] = list(
            dict.fromkeys(itertools.chain.from_iterable(ordered_plugins))
        )

        # Ensure all referenced plugins are defined and valid
        for plugin in parsed["environments"][environment]:
            if plugin in ConfigSchema().fields:
                raise ConfigurationError(
                    f"Configuration error: environment {environment} references reserved name {plugin}"
                )
            if plugin not in parsed:
                raise ConfigurationError(
                    f"Configuration error: environment {environment} references undefined plugin {plugin}"
                )

    return parsed


def from_file(config_file=None) -> Configuration:
    if not config_file:
        config_file = os.environ.get("ZOCALO_CONFIG")
    if not config_file:
        return Configuration({})
    config_file = pathlib.Path(config_file)
    return Configuration(_merge_configuration(None, config_file, config_file.parent))


def from_string(configuration: str) -> Configuration:
    return Configuration(_merge_configuration(configuration, None, pathlib.Path.cwd()))



from __future__ import annotations

import argparse
from typing import Iterable, List

__all__ = ["get_specified_environments"]


class _EnvParser(argparse.ArgumentParser):
    def __init__(self, *args, add_help=False, **kwargs):
        super().__init__(*args, add_help=add_help, **kwargs)

    def error(self, *args, **kwargs):
        # don't exit on error
        return None


def get_specified_environments(
    argv=None, *, arguments: Iterable[str] = ("-e", "--environment")
) -> List[str]:
    """Extract a list of all environments putatively specified on the command line.
    Returns an empty list if there are any parsing issues or there are no specified
    environments. This function is designed to be used as the first of two passes
    for parsing command line options, so that the environments can be taken into
    account when parsing all other commands."""

    env_parser = _EnvParser()
    env_parser.add_argument(
        *arguments,
        dest="envs",
        action="append",
        default=[],
    )
    selected_environments = env_parser.parse_known_args()
    if not selected_environments:
        return []

    return selected_environments[0].envs



from __future__ import annotations

import logging
from unittest import mock

import pytest

import zocalo.configuration
import zocalo.configuration.plugin_logging

sample_configuration = """
version: 1

logsetup:
  plugin: logging

  handlers:
    graylog:
      (): zocalo.configuration.plugin_logging.GraylogUDPHandler
      host: example.com
      port: 1234

  root:
    level: WARNING
    handlers: [ graylog ]

  loggers:
    zocalo:
      level: INFO
    dials:
      level: INFO
    pika:
      level: WARNING

  verbose:
    - loggers:
        dials:
          level: DEBUG
        pika:
          level: INFO

    - loggers:
        pika:
          level: DEBUG

environments:
  live:
    - logsetup
"""


@mock.patch("zocalo.configuration.plugin_logging.logging")
def test_plugin_sets_up_logging_with_variable_verbosity(logging):
    zc = zocalo.configuration.from_string(sample_configuration)
    logging.config.dictConfig.assert_not_called()
    assert zc.logging is None

    zc.activate_environment("live")
    logging.config.dictConfig.assert_called_once_with(
        {
            "disable_existing_loggers": False,
            "handlers": {
                "graylog": {
                    "()": "zocalo.configuration.plugin_logging.GraylogUDPHandler",
                    "host": "example.com",
                    "port": 1234,
                }
            },
            "incremental": False,
            "loggers": {
                "dials": {"level": "INFO"},
                "pika": {"level": "WARNING"},
                "zocalo": {"level": "INFO"},
            },
            "root": {"handlers": ["graylog"], "level": "WARNING"},
            "verbose": mock.ANY,
            "version": 1,
        }
    )
    assert zc.logging.verbosity == 0

    logging.config.reset_mock()
    logging.config.dictConfig.assert_not_called()

    zc.logging.verbosity = 1

    logging.config.dictConfig.assert_called_once_with(
        {
            "incremental": True,
            "loggers": {"dials": {"level": "DEBUG"}, "pika": {"level": "INFO"}},
            "version": 1,
        }
    )
    assert zc.logging.verbosity == 1

    logging.config.reset_mock()
    logging.config.dictConfig.assert_not_called()

    # Verbosity can't be reduced
    zc.logging.verbosity = 0
    assert zc.logging.verbosity == 1
    logging.config.dictConfig.assert_not_called()

    zc.logging.verbosity = 1
    assert zc.logging.verbosity == 1
    logging.config.dictConfig.assert_not_called()

    # Verbosity can't be increased beyond maximum
    zc.logging.verbosity = 7
    logging.config.dictConfig.assert_called_once_with(
        {"incremental": True, "loggers": {"pika": {"level": "DEBUG"}}, "version": 1}
    )
    assert zc.logging.verbosity == 2


@pytest.mark.parametrize("protocol", ("UDP", "TCP"))
@mock.patch("zocalo.configuration.plugin_logging.graypy")
@mock.patch("zocalo.configuration.plugin_logging.graypy.handler")
@mock.patch("zocalo.configuration.plugin_logging.socket")
def test_graypy_handlers_are_set_up_correctly(socket, handler, graypy, protocol):
    graypy_handler = getattr(graypy, f"GELF{protocol}Handler")
    zocalo_handler = getattr(
        zocalo.configuration.plugin_logging, f"Graylog{protocol}Handler"
    )

    graypy_handler.assert_not_called()
    socket.gethostbyname.assert_not_called()

    h = zocalo_handler(host=mock.sentinel.host, port=mock.sentinel.port)

    socket.gethostbyname.assert_called_once_with(mock.sentinel.host)
    graypy_handler.assert_called_once_with(
        socket.gethostbyname.return_value, mock.sentinel.port, level_names=True
    )
    assert h == graypy_handler.return_value


@mock.patch("zocalo.configuration.plugin_logging.logging")
def test_incremental_configuration_with_handler_definition_is_rejected(logging):
    zc = zocalo.configuration.from_string(
        """
version: 1

logsetup:
  plugin: logging

  handlers:
    graylog:
      (): zocalo.configuration.plugin_logging.GraylogUDPHandler
      host: example.com
      port: 1234

  incremental: True

  root:
    level: WARNING
    handlers: [ graylog ]

environments:
  live:
    - logsetup
"""
    )
    with pytest.raises(zocalo.ConfigurationError, match="incremental"):
        zc.activate_environment("live")


@mock.patch("zocalo.configuration.plugin_logging.logging")
def test_configuration_with_incremental_verbosity_definition_and_handler_is_rejected(
    logging,
):
    zc = zocalo.configuration.from_string(
        """
version: 1

logsetup:
  plugin: logging

  handlers:
    graylog:
      (): zocalo.configuration.plugin_logging.GraylogUDPHandler
      host: example.com
      port: 1234

  verbose:
    - root:
        level: WARNING
        handlers: [ graylog ]

environments:
  live:
    - logsetup
"""
    )
    with pytest.raises(zocalo.ConfigurationError, match="verbosity.*incremental"):
        zc.activate_environment("live")


@mock.patch("zocalo.configuration.plugin_logging.logging")
def test_configuration_with_invalid_verbosity_definitions_is_rejected(logging):
    zc = zocalo.configuration.from_string(
        """
version: 1

logsetup:
  plugin: logging

  verbose:
    - DEBUG

environments:
  live:
    - logsetup
"""
    )
    with pytest.raises(zocalo.ConfigurationError, match="not a dictionary"):
        zc.activate_environment("live")


def test_downgrade_filter_downgrades_log_messages():
    record = logging.LogRecord(
        "some.logger", logging.ERROR, "path", 10, "msg", (), None
    )

    f = zocalo.configuration.plugin_logging.DowngradeFilter(
        "WARNING", only_below="CRITICAL"
    )
    f.filter(record)

    assert record.levelname == "WARNING"
    assert record.levelno == logging.WARNING


def test_downgrade_filter_leaves_low_level_messages_alone():
    record = logging.LogRecord("some.logger", logging.INFO, "path", 10, "msg", (), None)

    f = zocalo.configuration.plugin_logging.DowngradeFilter(
        "WARNING", only_below="CRITICAL"
    )
    f.filter(record)

    assert record.levelname == "INFO"
    assert record.levelno == logging.INFO


def test_downgrade_filter_leaves_high_level_messages_alone():
    record = logging.LogRecord(
        "some.logger", logging.CRITICAL, "path", 10, "msg", (), None
    )

    f = zocalo.configuration.plugin_logging.DowngradeFilter(
        "WARNING", only_below="CRITICAL"
    )
    f.filter(record)

    assert record.levelname == "CRITICAL"
    assert record.levelno == logging.CRITICAL



from __future__ import annotations

import logging
import socket
import warnings

import graypy.handler
from marshmallow import fields, validate

from zocalo.configuration import PluginSchema


class _PythonLevelToSyslogConverter:
    """
    A helper class to monkeypatch graypy to support log levels in graylog.
    This translates Python integer level numbers to syslog levels.
    """

    @staticmethod
    def get(level, _):
        if level < 20:
            return 7  # DEBUG
        elif level < 25:
            return 6  # INFO
        elif level < 30:
            return 5  # NOTICE
        elif level < 40:
            return 4  # WARNING
        elif level < 50:
            return 3  # ERROR
        elif level < 60:
            return 2  # CRITICAL
        else:
            return 1  # ALERT


class Graylog:
    """
    A plugin to enable logging to a Graylog server using graypy.
    """

    class Schema(PluginSchema):
        protocol = fields.Str(validate=validate.OneOf(["UDP", "TCP"]), required=True)
        host = fields.Str(required=True)
        port = fields.Int(required=True)

    @staticmethod
    def activate(configuration):
        warnings.warn(
            "The Graylog configuration plugin will soon be deprecated."
            " It will be replaced by the more powerful logging plugin.",
            PendingDeprecationWarning,
        )

        graypy.handler.SYSLOG_LEVELS = _PythonLevelToSyslogConverter()

        # Create and enable graylog handler
        if configuration["protocol"] == "UDP":
            handler = graypy.GELFUDPHandler
        else:
            handler = graypy.GELFTCPHandler
        host = configuration["host"]
        try:
            # Attempt to look up the hostname only once during setup
            host = socket.gethostbyname(host)
        except Exception:
            pass
        graylog = handler(host, configuration["port"], level_names=True)
        logger = logging.getLogger()
        logger.addHandler(graylog)

        # Return the handler, which may be useful to attach filters to it.
        return graylog
